# Individual checks that were once contained in individual scripts.
# Moved here to reduce clutter in base directory.

package LARRD::checks;

use strict;
use FastGlob qw(glob);
use FileHandle;
use RRDs 1.000411; # make site-perl-install when installing
use lib defined($ENV{'LARRDHOME'}) ? "$ENV{'LARRDHOME'}/lib" : '/usr/local/larrd/lib';
use LARRD::graphinfo qw( &get_dsnames );

use Exporter;

use vars qw (@ISA @EXPORT @EXPORT_OK);

@ISA = qw(Exporter);

@EXPORT = qw(  );

@EXPORT_OK = qw( %larrdchecks );

our %larrdchecks = ();
$larrdchecks{'env'} = \&env_check;
$larrdchecks{'apache'} = \&apache_check;
$larrdchecks{'bbnet'} = \&bbnet_check;
$larrdchecks{'bind'} = \&bind_check;
$larrdchecks{'disk'} = \&disk_check;
$larrdchecks{'inode'} = \&inode_check;
$larrdchecks{'iostat'} = \&iostat_check;
$larrdchecks{'procs'} = \&procs_check;
$larrdchecks{'users'} = \&users_check;
$larrdchecks{'la'} = \&la_check;
$larrdchecks{'memory'} = \&memory_check;
$larrdchecks{'netstat'} = \&netstat_check;
$larrdchecks{'sendmail'} = \&sendmail_check;
$larrdchecks{'vmstat'} = \&vmstat_check;
$larrdchecks{'mailq'} = \&mailq_check;
$larrdchecks{'temperature'} = \&temperature_check;
$larrdchecks{'files'} = \&files_check;
$larrdchecks{'ftpusers'} = \&ftpusers_check;

#--------------------
# Prevent stupid typos with cruel autoload!
#--------------------

sub AUTOLOAD {
  use vars qw($AUTOLOAD);
  print("package ".(caller(1))[0].", file ".(caller(1))[1].", line ".(caller(1))[2].", subname ".(caller(1))[3].", called the following function which doesn't exist: $AUTOLOAD\n");
  exit;
}


#----
# &unlink_data($log);
#
# Remove data files after they have been processed
#----

sub unlink_data {
  my($log) = @_;
  if ($log =~ /$ENV{'BBVAR'}\/data/io) {
    unless (unlink($log)) {
      print(localtime().": ".(caller(1))[3].": ERROR could not unlink $log: $!\n");
    }
  } # if ($log =~ /$ENV{'BBVAR'}\/data/io)
  return 1;
}

#---
# &env_check();
#
# Check environment variables
#----

sub env_check {
  if (not $ENV{'BBHOME'}) {
    print(localtime().": ".(caller(0))[3].": BBHOME is not set\n");
    return;
  }
  if (not -d $ENV{'BBHOME'}) {
    print(localtime().": ".(caller(0))[3].": BBHOME is invalid\n");
    return;
  }
  if (not $ENV{'BBRRDS'}) {
    print(localtime().": ".(caller(0))[3].": BBRRDS is not set\n");
    return;
  }
  if (not -d $ENV{'BBRRDS'}) {
    if (mkdir "$ENV{'BBRRDS'}",0755) {
      print(localtime().": ".(caller(0))[3].": WARN did not find $ENV{'BBRRDS'}, created. This is perfectly normal if larrd is being run for the first time.\n");
    } else {
      print(localtime().": ".(caller(0))[3].": ERROR did not find $ENV{'BBRRDS'}, could not create: $!\n");
      return;
    }
  }
  return 1;
} # sub env_check


#----
# &apache_check();
#
# Check Apache server-status URLs
#----

sub apache_check {

  # Apache's DS's

  my @ds = ("DS:TA:DERIVE:600:0:U",   # Total Access
       "DS:TKB:DERIVE:600:0:U",  # Total kBytes
       "DS:BS:GAUGE:600:1:U",     # Busy Servers
       "DS:IS:GAUGE:600:1:U");    # Idle Servers

  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.apache"),&glob("$ENV{'BBVAR'}/data/*.apache")) {
    (my $commahost = $log) =~ s/^.*\/|\.apache$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $log;
  }
  while (my($host,$log) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;
    my $ctime = (stat($log))[10];

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log '$log': $!");
      next;
    }

    my($stuff,$ta,$tkb,$bs,$is);
    while ( defined (my $status = $infh->getline()) ) {
      chomp $status;
      print(localtime().": ".(caller(0))[3].": DEBUG $status\n") if $ENV{'DEBUG'};

      if($status =~ /Total Accesses/) {
        ($stuff,$ta) = split ": ",$status;
        next;
      } elsif ($status =~ /Total kBytes/) {
        ($stuff,$tkb) = split ": ",$status;
        next;
      } elsif ($status =~ /BusyServers/) {
        ($stuff,$bs) = split ": ",$status;
        next;
      } elsif ($status =~ /IdleServers/) {
        ($stuff,$is) = split ": ",$status;
        next;
      } # if($status =~ /Total Accesses/)
    } # while ( defined ($status = <$socket>) )
    $infh->close;
    unless (defined($ta) && defined($tkb) && defined($bs) && defined($is)) {
      print(localtime().": ".(caller(0))[3].": WARN Apache on $host has ExtendedStatus set to OFF\n");
    }

    print(localtime().": ".(caller(0))[3].": DEBUG TA=$ta, TKB=$tkb, BS=$bs, IS=$is\n") if ($ENV{'DEBUG'}); 

    my $RRD="$ENV{'BBRRDS'}/$commahost.apache.rrd";
    my $oldRRD="$ENV{'BBRRDS'}/$host.apache.rrd";
    if (($RRD ne $oldRRD) &&
        (-f $oldRRD)) {
      unless (rename($oldRRD,$RRD)) {
        print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
      }
    }

    print(localtime().": ".(caller(0))[3].": RRD=$RRD\n") if ($ENV{'DEBUG'});

    # see if the RRD is around, if not make it  
    if ( not -f $RRD ) {
    
      # but thought 'standardizing' RRD's would be worth it
      # 30 seconds....
      my @rras = split " ",$ENV{'RRAS'};
      RRDs::create($RRD,@ds,@rras);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
        next;
      }
      print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
      $ctime = 'N';
    } else {
      if ($ctime == RRDs::last($RRD)) {
        print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
        next;
      }
    }

    # Check counter last values to make sure counter isn't less that
    # previous.  It that is the case insert a U into rrdtool.
    # We use the info last_ds.

    my ($info) = RRDs::info($RRD);
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": ERROR could not fetch last update for $RRD: ".RRDs::error."\n");
    }

    print(localtime().": ".(caller(0))[3].": DEBUG ta_last($info->{'ds[TA].last_ds'}) ta ($ta) tkb_last($info->{'ds[TKB].last_ds'}) tkb ($tkb)\n") if ($ENV{'DEBUG'});

    # now update the RRD
    RRDs::update("$RRD","--template","TA:TKB:BS:IS","$ctime:$ta:$tkb:$bs:$is");
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
    }
    &unlink_data($log);
  } # while (my($host,$log) = each(%host2log))
  return 1;
} # sub apache_check


#----
# &bbnet_check();
#
# Check all BB checked network services except http
#----

sub bbnet_check {

  # http will be tougher cause log can check multiple sites
  foreach my $bbnet ("ssh","telnet","nntp","ftp","pop","pop-2","pop2","pop-3","pop3","smtp","imap2","conn","fping") {
    foreach my $log (&glob("$ENV{'BBLOGS'}/*.$bbnet")) {
      (my $commahost = $log) =~ s/^.*\/|\.$bbnet$//g;
      (my $host = $commahost) =~ s/\,/\./g;
      print(localtime().": ".(caller(0))[3].": Log file $log\n") if $ENV{'DEBUG'};
    
      my $RRD="$ENV{'BBRRDS'}/$commahost.bbnet.$bbnet.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.bbnet.$bbnet.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }
      print(localtime().": ".(caller(0))[3].": RRD File $RRD\n") if $ENV{'DEBUG'};
      # Move old one into place if it is there
      if (-f "$ENV{'BBRRDS'}/$commahost.tcp.$bbnet.rrd") {
        rename("$ENV{'BBRRDS'}/$commahost.tcp.$bbnet.rrd",$RRD);
      }
    
      # now lets get the data 
      # look through the file and get what you want
      # hacked in the conn since it varies from bbnet, but trends tcp connects
    
      my $ctime = (stat($log))[10];
      # Skip if I already got data from this file
      my $infh = new FileHandle;

      unless ($infh->open("< $log")) {
        print(localtime().": ".(caller(0))[3].": WARN Unable to open log '$log': $!");
        next;
      }
      my @statlines = $infh->getlines();
      $infh->close;
      chomp(@statlines);
    
      if ($statlines[0] =~ /OK: SERVICE IS NOT RUNNING|^clear .* \(dialup\)$/) {
        # It shouldn't be running. Forget I ever saw it.
        print(localtime().": ".(caller(0))[3].": DEBUG $host $bbnet should not be running or is happily not running. $log skipped.\n") if $ENV{'DEBUG'};
        next;
      }
      if ((grep(/DATA TRUNCATED/,@statlines)) && (! grep(/sec/,@statlines))) {
        print(localtime().": ".(caller(0))[3].": DEBUG $host $bbnet shows a hung connection. $log skipped.\n") if $ENV{'DEBUG'};
        next;
      }
      if ($statlines[0] =~ /^clear /) {
        print(localtime().": ".(caller(0))[3].": DEBUG $host $bbnet is showing status clear. $log skipped.\n") if $ENV{'DEBUG'};
        next;
      }
      if ($statlines[0] =~ /^blue /) {
        print(localtime().": ".(caller(0))[3].": DEBUG $host $bbnet is showing status blue. $log skipped.\n") if $ENV{'DEBUG'};
        next;
      }
      if ($statlines[0] !~ /green|yellow|red/) {
        print(localtime().": ".(caller(0))[3].": WARN $host $bbnet is showing status '$statlines[0]', not red, yellow, or green. $log skipped.\n");
        next;
      }
      if ($statlines[0] =~ /^red /) {
        # Service is down. Can't get useful data.
        print(localtime().": ".(caller(0))[3].": DEBUG $host $bbnet is down. $log skipped.\n") if $ENV{'DEBUG'};
        next;
      }
      if ($statlines[0] =~ /^blue /) {
        # Service is offline. Can't get useful data.
        print(localtime().": ".(caller(0))[3].": DEBUG $host $bbnet is offline. $log skipped.\n") if $ENV{'DEBUG'};
        next;
      }
    
      # fping
    
      my $sec = undef;
      foreach (my $n = 0; $n <= $#statlines; $n++) {
        my $line = $statlines[$n];
        my $dummy;
        if(($line =~ /Seconds/) && ($bbnet ne 'conn')) {
          ($dummy,$sec) = split " ",$line,3;
          last;
        } elsif (($line =~ /time=/) && ($bbnet eq 'conn')) {
          # is this the best method for finding ping conn?
          ($dummy,$sec) = split "time=",$line;
          ($sec,$dummy) = split " ",$sec;
          $sec=$sec/1000; # convert milliseconds to seconds
          if($dummy =~ /^usec/) {
            $sec=$sec/1000;
          }
          if($dummy =~ /^sec/) {
            $sec=$sec*1000;
          }
        } elsif (($line =~ /alive/) && (($bbnet eq 'fping') || ($bbnet eq 'conn'))) {
          if ($line =~ /\(([\d.]+)\s+ms\)/) {
            $sec = $1;
            $sec = $sec / 1000;
          }
        }     
      }
      print(localtime().": ".(caller(0))[3].": DEBUG sec: $sec\n") if $ENV{'DEBUG'}; 
      if (defined($sec)) {
      
        # If it is looking at the log with no larrdable data
        # it won't create the rrd and in turn the graph
        # it makes sense this way.
    
        # if it doesn't exist, make it
        # If it does, skip it if it's already been updated
        if (-f $RRD) {
          if ($ctime == RRDs::last($RRD)) {
            print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
            next;
          }
        } else {
          # RRAS defined in larrd.pl
          my @rras = split(" ",$ENV{'RRAS'});
          RRDs::create($RRD,"DS:sec:GAUGE:600:0:U",@rras);
          if (RRDs::error) {
            print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
            next;
          }
          $ctime="N";
          print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        } # if (-f $RRD )
    
        print(localtime().": ".(caller(0))[3].": DEBUG update $RRD $ctime:$sec\n") if ($ENV{'DEBUG'});
        RRDs::update($RRD,"--template","sec","$ctime:$sec");
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": WARN updating $RRD from log file $log: ".RRDs::error."\n");
        }
      } else { # if ($sec ne "")
        # Too many false alarms on this one
        print(localtime().": ".(caller(0))[3].": DEBUG No indication of time elapsed found in $log, status '$statlines[0]'\n") if $ENV{'DEBUG'};
      } # if ($sec ne "")
    } # foreach my $log (glob("$ENV{'BBLOGS'}/*.$bbnet"))
  } # foreach my $bbnet

  # HTTP check
  unless (&http_check()) {
    print(localtime().": ".(caller(0))[3].": ERROR http check failed.\n");
    return;
  }

  return 1;
} # sub bbnet_check


#----
# &http_check();
#
# Different from the Apache check.
# The Apache check measures performance of Apache httpd servers.
# This checks whether or not any kind of server is alive.
#----

sub http_check {
  
  # Now we will do http since it have have several urls we break up the
  # rrds into seperate files for a 1 to 1 relationship.  This allows
  # easy additions of urls.
  
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.http")) {
    (my $commahost = $log) =~ s/^.*\/|\.http$//g;
    (my $host = $commahost) =~ s/\,/\./g;

    print(localtime().": ".(caller(0))[3].": Found log file $log for http\n") if $ENV{'DEBUG'};
    # you need three things to collect data
    # RRD - an RRD
    # ctime  - a timestamp (in epoch)
    # sec - value of data (could be multiple values, see disk-larrd.sh for example)
  
    my $ctime = (stat($log))[10];
  
    # Read the http urls in to a hash
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": WARN Unable to open log '$log': $!");
      next;
    }
  
    my $status = $infh->getline();
  
    if ($status !~ /green|yellow|red/) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing red, yellow, or green. $log skipped.\n");
      next;
    }
  
    my %http = ();
    my $url = undef;
    while (defined (my $line = $infh->getline())){
      my $dummy = undef;
      if($line =~ /^&/) {
        ($dummy,$url) = split " ",$line,3;
        $url =~ s/http\:\/\///g;
        $url =~ s/\//\,/g;
      }
      if($line =~ /Seconds/) {
        my $sec = undef;
        ($dummy,$sec) = split " ",$line,3;
        $http{$url}=$sec;
      }
    }
    
    # with the hash in place we will roll our own rrds
    while (my($url,$sec) = each(%http)) {
      my $RRD="$ENV{'BBRRDS'}/$commahost.bbnet.http.$url.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.bbnet.http.$url.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }
      print(localtime().": ".(caller(0))[3].": RRD File $RRD\n") if $ENV{'DEBUG'};
  
      if ($sec ne "") {
        # if it doesn't exist, make it
        if ( not -f $RRD ) {
          # RRAS defined in larrd.pl
          my @rras = split(" ",$ENV{'RRAS'});
          RRDs::create($RRD,"DS:sec:GAUGE:600:0:U",@rras);
          if (RRDs::error) {
            print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
            next;
          }
          $ctime="N";
          print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        } else { #  if ( not -f $RRD )
          if ($ctime == RRDs::last($RRD)) {
            print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
            next;
          }
        } # if ( not -f $RRD )
  
        RRDs::update($RRD,'--template','sec',"$ctime:$http{$url}");
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
        }
      } # if ($sec ne "")
    } # foreach $url (keys %http)
  } # while (my($host,$log) = each(%host2log))
  return 1;
} # sub http_check


#----
# &bind_check();
#
# Check bind status
#----

sub bind_check {

  # add statistics-intervel 5; to your options in named.conf if you run bind
  #

  # Bind DS's

  my @ds = ("DS:A:DERIVE:600:0:U",    # network address
            "DS:ANY:DERIVE:600:0:U",  # all/any information
            "DS:AXFR:DERIVE:600:0:U",
            "DS:MX:DERIVE:600:0:U",   # mail exchanger
            "DS:O:DERIVE:600:0:U",    # ?
            "DS:PTR:DERIVE:600:0:U",  # ?
            "DS:SOA:DERIVE:600:0:U"); # zone of authority record

  my %bind_type = ("A"   =>1,
              "ANY" =>1,
              "AXFR"=>1,
              "MX"  =>1,
              "O"   =>1,
              "PTR" =>1,
              "SOA" =>1);

  # NOTE : See README about $ENV{'LARRDCOMM'}

  # $ENV{'LARRDCOMM'} has gone away.  Only the bottom feeders
  # specify how the data gets there.  The larrd drivers look in
  # the logs directory first.  Then they look in the data directory.
  # Data logs take precedence.

  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.bind"),&glob("$ENV{'BBVAR'}/data/*.bind")) {
    (my $commahost = $log) =~ s/^.*\/|\.bind$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $log;
  }
  while (my($host,$log) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;

    print(localtime().": ".(caller(0))[3].": DEBUG log=$log\n") if ($ENV{'DEBUG'});
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": WARN Unable to open log '$log': $!");
      next;
    }
    my $temp = $infh->getline() if $log =~ /$ENV{'BBLOGS'}/io;
    my $bind = $infh->getline();
    print(localtime().": ".(caller(0))[3].": DEBUG bind = $bind\n") if ($ENV{'DEBUG'});

    $infh->close;
    chomp $bind;

    my($stats,$ctime,@stats);
    (undef,$stats) = split "]: ",$bind;
    (undef,$ctime,undef,@stats) = split " ",$stats;

    my(%stats_h);
    foreach (@stats) {
      my($type,$value) = split "=";

      print(localtime().": ".(caller(0))[3].": DEBUG type = $type, value = $value\n") if ($ENV{'DEBUG'});
      $stats_h{$type}=$value;
    }

    my $RRD="$ENV{'BBRRDS'}/$commahost.bind.rrd";
    my $oldRRD="$ENV{'BBRRDS'}/$host.bind.rrd";
    if (($RRD ne $oldRRD) &&
        (-f $oldRRD)) {
      unless (rename($oldRRD,$RRD)) {
        print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
      }
    }

    print(localtime().": ".(caller(0))[3].": DEBUG RRD = $RRD\n") if ($ENV{'DEBUG'});
    print(localtime().": ".(caller(0))[3].": DEBUG ds = ".join(',',@ds)."\n") if ($ENV{'DEBUG'});

    # see if the RRD is around, if not make it  
    if ( not -f $RRD ) {
    
      # but thought 'standardizing' RRD's would be worth it
      # 30 seconds....
      my @rras = split " ",$ENV{'RRAS'};
      RRDs::create($RRD,"--start","$ctime-1",@ds,@rras);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
        next; #does this work? yup!
      }
      $ctime="N";
      print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
    } # if ( not -f $RRD )

    my($rrdupdate_t);
    foreach my $col ( sort keys %bind_type) { 
      $rrdupdate_t = "$rrdupdate_t$col:";
    }
    chop $rrdupdate_t;

    print(localtime().": ".(caller(0))[3].": DEBUG rrdupdate_t = $rrdupdate_t\n") if ($ENV{'DEBUG'});

    my($rrdupdate) = $ctime;
    foreach my $col ( sort keys %bind_type) { 
      $stats_h{$col} = 0 unless $stats_h{$col};
      $rrdupdate = "$rrdupdate:$stats_h{$col}";
    }

    print(localtime().": ".(caller(0))[3].": DEBUG rrdupdate = $rrdupdate\n") if ($ENV{'DEBUG'});

    # now update the RRD
    RRDs::update("$RRD","-t","$rrdupdate_t","$rrdupdate");
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
    }
    &unlink_data($log);
  } # while (my($host,$log) = each(%host2log))
  return 1;
} # sub bind_check


#----
# &disk_check()
#
# Check Disk logs
#----

sub disk_check {

  foreach my $log (&glob("$ENV{'BBLOGS'}/*.disk")) {
    (my $commahost = $log) =~ s/^.*\/|\.disk$//g;
    (my $host = $commahost) =~ s/\,/\./g;

    # Slurp the log into an array for now and later use
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Couldn't open log file $log: $!\n");
      next;
    }
    my @log = $infh->getlines();
    $infh->close;
  
    my $ctime = (stat($log))[10];
    print(localtime().": ".(caller(0))[3].": DEBUG Timestamp of '$log' is ".localtime($ctime)."\n") if $ENV{'DEBUG'};

    if ($log[0] !~ /clear|green|yellow|red/) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing clear, red, yellow, or green. $log skipped.\n");
      next;
    }

    # Ugly I know, but effiecent
    my $whole = join "",@log;

    # Decide the client system type
    my $systemtype = undef;
    if ($whole =~ / xfs | efs | cxfs / ) {
      $systemtype="IRIX";
    } elsif ($whole =~ /DASD/ ) { 
      $systemtype="AS400";
    } elsif ($whole =~ /\/FIXED/ ) {
      $systemtype="NT";
    } else {
      $systemtype="UNIX";
    }

    # Pop the first line. I don't seem to use it, and if I don't shift it
    # I might get a disk named /red or something if I'm using NT.
    shift(@log);

    foreach my $line (@log) {
      next if ($line =~ /^Filesystem .* Mounted on$/);
      next if ($line =~/^&/);
      next if ($line !~/\// && $systemtype ne "AS400" );
      next if ($line !~/DASD/ && $systemtype eq "AS400" );
      next if ($line =~/^\s+(yellow|red)\s+\//);
      last if ($line =~/bloater/);

      print(localtime().": ".(caller(0))[3].": DEBUG $line\n") if $ENV{'DEBUG'};

      my @col = split " ",$line;

      # sometimes NT is weird
      # what am i'm saying it's always weird.
      # can i waste any more CPU cycles?
      my $diskname = undef;
      if ( $systemtype eq "AS400" ) {
        $diskname=",DASD"; 
      } elsif ( $systemtype eq "NT" ) {
        $diskname="\,$col[0]";
      } elsif ( $systemtype eq "IRIX" ) {
        # /'s are a nasty thing to have in filenames
        $col[6] =~ s/\//\,/g;
        $diskname=$col[6];
      } else {
        # /'s are a nasty thing to have in filenames
        $col[5] =~ s/\//\,/g;
        $diskname=$col[5]; 
      }

      # want / filesytems to have nice name
      if ( $diskname eq "\," ) {
        $diskname="\,root"
      }
  
      my $RRD="$ENV{'BBRRDS'}/$commahost.disk$diskname.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.disk$diskname.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }
  
      # if it doesn't exist, make it
      if (-f $RRD) {
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        }
      } else {
        # RRAS defined in larrd.pl
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"DS:pct:GAUGE:600:0:100","DS:used:GAUGE:600:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        # New RRDs don't like old dates
        $ctime="N";
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'}); 
      }

      # going to get % cause it is relative
      my $pct = undef;
      if ( $systemtype eq "AS400" ) {
        $col[12] =~ s/%//g;
        $pct=$col[12];
      } elsif ( $systemtype eq "IRIX" ) {
        $col[5] =~ s/%//g;
        $pct=$col[5];
      } else {
        $col[4] =~ s/%//g;
        $pct=$col[4];
      }

      # going to get used cause it is absolute
      my $used = undef;
      if ( $systemtype eq "AS400" ) {
        $used=0;  
      } elsif ( $systemtype eq "IRIX" ) {
        $used=$col[3];
      } else {
        $used=$col[2];
      }
  
      print("$RRD $ctime:$pct:$used\n") if ($ENV{'DEBUG'});

      RRDs::update("$RRD",'--template','pct:used',"$ctime:$pct:$used");
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }
    } # foreach $line (@log)
  } # while (  defined (my $log = <*.disk>)  )
  return 1;
} # sub disk_check


#----
# &inode_check()
#
# Check Disk inode logs
#----

sub inode_check {

  foreach my $log (&glob("$ENV{'BBLOGS'}/*.inode")) {
    (my $commahost = $log) =~ s/^.*\/|\.inode$//g;
    (my $host = $commahost) =~ s/\,/\./g;

    # Slurp the log into an array for now and later use
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Couldn't open log file $log: $!\n");
      next;
    }
    my @log = $infh->getlines();
    $infh->close;
  
    my $ctime = (stat($log))[10];
    print(localtime().": ".(caller(0))[3].": DEBUG Timestamp of '$log' is ".localtime($ctime)."\n") if $ENV{'DEBUG'};

    if ($log[0] !~ /clear|green|yellow|red/) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing clear, red, yellow, or green. $log skipped.\n");
      next;
    }

    # Ugly I know, but effiecent
    my $whole = join "",@log;

    # Decide the client system type
    my $systemtype = undef;
    if ($whole =~ / xfs | efs | cxfs / ) {
      $systemtype="IRIX";
    } elsif ($whole =~ /DASD/ ) { 
      $systemtype="AS400";
    } elsif ($whole =~ /\/FIXED/ ) {
      $systemtype="NT";
    } else {
      $systemtype="UNIX";
    }

    # Pop the first line. I don't seem to use it, and if I don't shift it
    # I might get a disk named /red or something if I'm using NT.
    shift(@log);

    foreach my $line (@log) {
      next if ($line =~ /^Filesystem .* Mounted on$/);
      next if ($line =~/^&/);
      next if ($line !~/\// && $systemtype ne "AS400" );
      next if ($line !~/DASD/ && $systemtype eq "AS400" );
      next if ($line =~/^\s+(yellow|red)\s+\//);
      last if ($line =~/bloater/);

      print(localtime().": ".(caller(0))[3].": DEBUG $line\n") if $ENV{'DEBUG'};

      my @col = split " ",$line;

      # sometimes NT is weird
      # what am i'm saying it's always weird.
      # can i waste any more CPU cycles?
      my $diskname = undef;
      if ( $systemtype eq "AS400" ) {
        $diskname=",DASD"; 
      } elsif ( $systemtype eq "NT" ) {
        $diskname="\,$col[0]";
      } elsif ( $systemtype eq "IRIX" ) {
        # /'s are a nasty thing to have in filenames
        $col[6] =~ s/\//\,/g;
        $diskname=$col[6];
      } else {
        # /'s are a nasty thing to have in filenames
        $col[5] =~ s/\//\,/g;
        $diskname=$col[5]; 
      }

      # want / filesytems to have nice name
      if ( $diskname eq "\," ) {
        $diskname="\,root"
      }
  
      my $RRD="$ENV{'BBRRDS'}/$commahost.inode$diskname.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.inode$diskname.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }
  
      # if it doesn't exist, make it
      if (-f $RRD) {
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        }
      } else {
        # RRAS defined in larrd.pl
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"DS:pct:GAUGE:600:0:100","DS:used:GAUGE:600:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        # New RRDs don't like old dates
        $ctime="N";
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'}); 
      }

      # going to get % cause it is relative
      my $pct = undef;
      if ( $systemtype eq "AS400" ) {
        $col[12] =~ s/%//g;
        $pct=$col[12];
      } elsif ( $systemtype eq "IRIX" ) {
        $col[5] =~ s/%//g;
        $pct=$col[5];
      } else {
        $col[4] =~ s/%//g;
        $pct=$col[4];
      }

      # going to get used cause it is absolute
      my $used = undef;
      if ( $systemtype eq "AS400" ) {
        $used=0;  
      } elsif ( $systemtype eq "IRIX" ) {
        $used=$col[3];
      } else {
        $used=$col[2];
      }
  
      print("$RRD $ctime:$pct:$used\n") if ($ENV{'DEBUG'});

      RRDs::update("$RRD",'--template','pct:used',"$ctime:$pct:$used");
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }
    } # foreach $line (@log)
  } # while (  defined (my $log = <*.inode>)  )
  return 1;
} # sub inode_check


#----
# &iostat_check();
#
# Check iostat results.
# Solaris only.
#----

sub iostat_check {

  my @ds = ("DS:rs:GAUGE:600:0:U",
       "DS:ws:GAUGE:600:0:U",
       "DS:krs:GAUGE:600:0:U",
       "DS:kws:GAUGE:600:0:U",
       "DS:wait:GAUGE:600:0:U",
       "DS:actv:GAUGE:600:0:U",
       "DS:wsvc_t:GAUGE:600:0:U",
       "DS:asvc_t:GAUGE:600:0:U",
       "DS:w:GAUGE:600:0:U",
       "DS:b:GAUGE:600:0:U",
       "DS:sw:GAUGE:600:0:U",
       "DS:hw:GAUGE:600:0:U",
       "DS:trn:GAUGE:600:0:U",
       "DS:tot:GAUGE:600:0:U",);
  my $dstemplate = '';
  foreach my $dsource (@ds) {
    if ($dsource =~ /^DS\:([^\:]+)\:/) {
      $dstemplate .= "$1:";
    }
  }
  chop($dstemplate);

  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.iostat"),&glob("$ENV{'BBVAR'}/data/*.iostat")) {
    (my $commahost = $log) =~ s/^.*\/|\.iostat$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $log;
  }
  while (my($host,$log) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;

    print(localtime().": ".(caller(0))[3].": DEBUG log=$log\n") if ($ENV{'DEBUG'});
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR unable to open log file '$log': $!\n");
      next;
    }

    my %d_h;
    while (defined(my $iostat = $infh->getline())) {
      next if $iostat =~ /BEGINKEY/;
      last if $iostat =~ /ENDKEY/;
      chomp $iostat;
      my ($device,$part) = split " ",$iostat;
      $d_h{$device}=$part;
    }
    while (defined(my $iostat = $infh->getline())) {
      next if $iostat =~ /BEGINDATA/;
      last if $iostat =~ /ENDDATA/;
      next if $iostat =~ /device/;
      chomp $iostat;
      my($rs,$ws,$krs,$kws,$wait,$actv,$wsvc_t,$asvc_t,
      $w,$b,$sw,$hw,$trn,$tot,$device) = split " ",$iostat;

      # 3. grab the timpstamp for rrd. 

      my $ctime = (stat($log))[10];
      print(localtime().": ".(caller(0))[3].": DEBUG ctime = $ctime\n") if ($ENV{'DEBUG'});

      # 4. create the rrd filename

      (my $diskname = $d_h{$device}) =~ s/\//\,/g;

      if ( $diskname eq "\," ) {
        $diskname="\,root"
      }


      my $commahost = $host;
         $commahost =~ s/\./\,/g;
      my $RRD="$ENV{'BBRRDS'}/$commahost.iostat.$diskname.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.iostat$diskname.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }

      print(localtime().": ".(caller(0))[3].": RRD=$RRD\n") if ($ENV{'DEBUG'});

      # see if the RRD is around, if not make it  
      if ( not -f $RRD ) {
        # keeping this much disk data seems overkill
        # but thought 'standardizing' RRD's would be worth it
        # 30 seconds....
        $ctime--;
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"-b $ctime",@ds,@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        $ctime++;
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
      } else { # if ( not -f $RRD )
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        }
      }

      # 5. Put your data from step 2 to the rrd

      print(localtime().": ".(caller(0))[3].": rs=$rs, ws=$ws, krs=$krs, kws=$kws, wait=$wait, actv=$actv, wsvc_t=$wsvc_t, asvc_t=$asvc_t, w=$w, b=$b, sw=$sw, hw=$hw, trn=$trn, tot=$tot\n") if ($ENV{'DEBUG'});

      # now update the RRD
      RRDs::update("$RRD",'--template',"$dstemplate","$ctime:$rs:$ws:$krs:$kws:$wait:$actv:$wsvc_t:$asvc_t:$w:$b:$sw:$hw:$trn:$tot");
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }
    } # while( defined($iostat = $infh->getline()) )
    $infh->close;
    &unlink_data($log);
  }  # foreach my $log

  return 1;
} # sub iostat_check

#----
# &procs_check()
#
# Check number of processes
#----

sub procs_check {
  # Glob the cpu logs
  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu")) {
    (my $commahost = $log) =~ s/^.*\/|\.cpu$//g;
    (my $host = $commahost) =~ s/\,/\./g;

    my $ctime = (stat($log))[10];

    # Grab Load Avg line
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log file $log: $!");
      next;
    }

    my(@lines) = $infh->getlines();
    chomp(@lines);
    $infh->close;

    if ($lines[0] !~ /green|yellow|red/) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing red, yellow, or green. $log skipped.\n");
      next;
    }

    # Attempt to grab procs and users
    if ($lines[0] =~ /(\d+)\s+users,\s+(\d+)\s+procs/o) {
      my $procs=$2;
  
      print(localtime().": ".(caller(0))[3].": DEBUG procs: $procs\n") if $ENV{'DEBUG'};
  
      # RRD for procs
  
      my $RRD = "$ENV{'BBRRDS'}/$commahost.procs.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.procs.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }

      # if it doesn't exist, make it 
      if (-f $RRD ) {
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        } else {
          # make sure it has the right DS name. If not, change it.
          my $ds_ref = &get_dsnames($RRD);
          if ((exists($ds_ref->{'la'})) && (! exists($ds_ref->{'procs'}))) {
            RRDs::tune($RRD,'-r','la:procs'); # Data source rename
            if (RRDs::error) {
              print(localtime().": ".(caller(0))[3].": ERROR renaming data source 'la' to 'procs' in $RRD: ".RRDs::error."\n");
              next;
            }
          }
          RRDs::update("$RRD",'-t','procs',"$ctime:$procs");
        }
      } else {
        # RRAS defined in larrd.pl
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"DS:procs:GAUGE:600:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        RRDs::update("$RRD",'-t','procs',"N:$procs");
      }

      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }
    } elsif ($lines[2] =~ /% Processor Time:\s+([\.\d]+)/) {
      # Vladimir's custom BBSend VBScript doesn't do procs or users
    } # elsif ($lines[2] =~ /% Processor Time:\s+([\.\d]+)/)
  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu"))

  return 1;
} # sub procs_check

#----
# &users_check()
#
# Check number of users
#----

sub users_check {
  # Glob the la logs
  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu")) {
    (my $commahost = $log) =~ s/^.*\/|\.cpu$//g;
    (my $host = $commahost) =~ s/\,/\./g;

    my $ctime = (stat($log))[10];

    # Grab Load Avg line
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log file $log: $!");
      next;
    }

    my(@lines) = $infh->getlines();
    chomp(@lines);
    $infh->close;

    if ($lines[0] !~ /green|yellow|red/) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing red, yellow, or green. $log skipped.\n");
      next;
    }

    # Attempt to grab procs and users.
    if ($lines[0] =~ /(\d+)\s+users,\s+(\d+)\s+procs/o) {
      my $users=$1;
  
      print(localtime().": ".(caller(0))[3].": DEBUG users: $users\n") if $ENV{'DEBUG'};

      # RRD to users

      my $RRD="$ENV{'BBRRDS'}/$commahost.users.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.users.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }

      # if it doesn't exist, make it 
      if (-f $RRD ) {
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        } else {
          # make sure it has the right DS name. If not, change it.
          my $ds_ref = &get_dsnames($RRD);
          if ((exists($ds_ref->{'la'})) && (! exists($ds_ref->{'procs'}))) {
            RRDs::tune($RRD,'-r','la:users'); # Data source rename
            if (RRDs::error) {
              print(localtime().": ".(caller(0))[3].": ERROR renaming data source 'la' to 'users' in $RRD: ".RRDs::error."\n");
              next;
            }
          }
          RRDs::update("$RRD",'-t','users',"$ctime:$users");
        }
      } else {
        # RRAS defined in larrd.pl
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"DS:users:GAUGE:600:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        RRDs::update("$RRD",'-t','users',"N:$users");
      }

      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }

    } elsif ($lines[2] =~ /% Processor Time:\s+([\.\d]+)/) {
      # Vladimir's custom BBSend VBScript doesn't give number of users
    } # elsif ($lines[2] =~ /% Processor Time:\s+([\.\d]+)/)
  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu"))

  return 1;
} # sub users_check



#----
# &la_check()
#
# Check *.cpu logs for: CPU Usage AND Load Average.
# This sub is written from scratch, replaces la_check_old().
# Many kinds of CPU information can be contained in the *.cpu log files.
# This checker tries to be more inclusive.
# Check for the following:
#   - Coworker's NT Scripts that give Processor Time in percent.
#   - Routers that give CPU Usage in intervals ranging from 1 second to 15 minutes.
#     From bb-xsnmpcpu.pl
#   - Output from top that contains both load average and CPU usage.
#   - Output from basic bb-cpu.sh that contains just 5 minute load average.
#   - Output from bb-rup.sh that contains the same.
# No point in getting CPU percentages from top output, not reliable.
# Busy and Idle are for NetApps, and busy can be just graphed as CPU.
# TODO: Graph percentage differently from load average.
# TODO: Change the name of the RRD from *.la.rrd to *.cpu.rrd.
#       This will entail changes in html.pm and config.pm.
#       Change la_check to cpu_check at that time.
#----

sub la_check {
  # from bb-xsnmpcpu.pl
  my $n = 0;
  my %statusorder = (
    cpu => $n++,
    cpu1sec => $n++,
    cpu5sec => $n++,
    cpu1min => $n++,
    cpu5min => $n++,
    cpu15min => $n++,
    la1min => $n++,
    la5min => $n++,
    la15min => $n++,
  );

  # Glob through the *.cpu logs
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu")) {
    # For each log, get a hostname.
    (my $commahost = $log) =~ s/^.*\/|\.cpu$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    my $ctime = (stat($log))[10];
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log file $log: $!");
      next;
    }
    my(@lines) = $infh->getlines();
    chomp(@lines);
    $infh->close;

    my %cpudb = ();
    my $useds = 'la';
    my $heartbeat = 600;
    foreach my $line (@lines) {
      if ($line =~ /load=([\d\.]+)$/) {
        $cpudb{'la'} = $1;
      } elsif ($line =~ /load=(\d+)\%/) { # BB NT client
        $cpudb{'cpu'} = $1;
      } elsif ($line =~ /CPU Usage\s*\:\s*(\d+)\s*\%/) {
        $cpudb{'cpu'} = $1;
      } elsif ($line =~ /\%\s+Processor Time\s*\:\s*([\-\d\.]+)/) { # Coworker's VBScript
        $cpudb{'cpu'} = $1;
        # $useds = 'ProcessorTime'; # Make coworker's VBScript use the same ds name.
        #                           Graph labels will change but -DLARRD dispay will work
        $heartbeat = 900; # Darned script runs only once every ten minutes
      } elsif ($line =~ /CPU Time in Busy Mode\:\s+(\d+)\%/) { # bb-xsnmpcpu.pl, NetApp
        $cpudb{'cpu'} = $1;
      } elsif ($line =~ /CPU Usage\, 1 Second Average\s*\:\s*(\d+)\s*\%/) { # bb-xsnmpcpu.pl
        $cpudb{'cpu1sec'} = $1;
      } elsif ($line =~ /CPU Usage\, 5 Second Average\s*\:\s*(\d+)\s*\%/) { # bb-xsnmpcpu.pl
        $cpudb{'cpu5sec'} = $1;
      } elsif ($line =~ /CPU Usage\, 1 Minute Average\s*\:\s*(\d+)\s*\%/) { # bb-xsnmpcpu.pl
        $cpudb{'cpu1min'} = $1;
      } elsif ($line =~ /CPU Usage\, 5 Minute Average\s*\:\s*(\d+)\s*\%/) { # bb-xsnmpcpu.pl
        $cpudb{'cpu5min'} = $1;
      } elsif ($line =~ /CPU Usage\, 15 Minute Average\s*\:\s*(\d+)\s*\%/) { # bb-xsnmpcpu.pl
        $cpudb{'cpu15min'} = $1;
      } elsif ($line =~ /Load Average, 1 Minute Average\s*\:\s*([\d\.]+)/) { # bb-rup.pl
        $cpudb{'la1min'} = $1;
      } elsif ($line =~ /Load Average, 5 Minute Average\s*\:\s*([\d\.]+)/) { # bb-rup.pl
        $cpudb{'la5min'} = $1;
      } elsif ($line =~ /Load Average, 15 Minute Average\s*\:\s*([\d\.]+)/) { # bb-rup.pl
        $cpudb{'la15min'} = $1;
      } elsif ($line =~ /load averages?\:\s+([\d\.]+)\,\s+([\d\.]+)\,\s+([\d\.]+)/) { # bb-cpu.sh with top
        ($cpudb{'la1min'},$cpudb{'la5min'},$cpudb{'la15min'}) = ($1,$2,$3);
      } elsif ($line =~ /Port mapper failure/) {
        # bb-rup.pl couldn't reach a down machine
        next;
      }
    }

    # Look through load average numbers.
    # Multiply by 100 if any of them have dots.
    # That's the BigBrother standard. :/ Well, less confusing this way.
    while (my($key,$value) = each(%cpudb)) {
      next unless $key =~ /^la/;
      if ($value =~ /\./) { $cpudb{$key} = $value * 100 }
    }

    # I only need the five minute average
    my $la = (exists($cpudb{'la5min'})) ? $cpudb{'la5min'} :
             (exists($cpudb{'la'})) ? $cpudb{'la'} :
             (exists($cpudb{'cpu5min'})) ? $cpudb{'cpu5min'} :
             (exists($cpudb{'cpu1min'})) ? $cpudb{'cpu1min'} :
             (exists($cpudb{'cpu'})) ? $cpudb{'cpu'} : undef;
    print(localtime().": ".(caller(0))[3].": DEBUG la: $la\n") if $ENV{'DEBUG'};
    unless (defined($la)) {
      print(localtime().": ".(caller(0))[3].": WARN no load average or cpu usage found in $log\n");
      next;
    }

    # Rename 0.42-style rrd files
    my $RRD="$ENV{'BBRRDS'}/$commahost.la.rrd";
    my $oldRRD="$ENV{'BBRRDS'}/$host.la.rrd";
    if (($RRD ne $oldRRD) &&
        (-f $oldRRD)) {
      unless (rename($oldRRD,$RRD)) {
        print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
      }
    }

    # if it doesn't exist, make it 
    if (-f $RRD ) {
      if ($ctime == RRDs::last($RRD)) {
        print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
        next;
      } else {
        # Figure out what DS I already have and use it or
        # change the name in the RRD
        my $ds_ref = &get_dsnames($RRD);
        if (! exists($ds_ref->{$useds})) {
          if (scalar(keys(%$ds_ref)) == 1) {
            # Change the name of the single DS
            my($existds) = keys(%$ds_ref);
            RRDs::tune($RRD,'-r',"$existds:$useds");
            if (RRDs::error) {
              print(localtime().": ".(caller(0))[3].": Unable to rename datasource in $RRD from $existds to $useds: ".RRDs::error."\n");
            }
          } else {
            # More than one DS in the RRD and I'm only trying to update one
            # variable. Oops.
            print(localtime().": ".(caller(0))[3].": Too many DS's in $RRD: ".join(',',keys(%$ds_ref))."\n");
          }
        }
      } # if ($ctime == RRDs::last($RRD))
    } else {
      # Create it
      $ctime = 'N';
      my @rras = split " ",$ENV{'RRAS'};
      RRDs::create($RRD,"DS:$useds:GAUGE:$heartbeat:0:U",@rras);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
        next;
      }
    } # if (-f $RRD )

    # Try to update
    RRDs::update($RRD,'-t',$useds,"$ctime:$la");
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      next; # No point since we're at the end but let's be consistent
    }

  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu"))

  return 1;
} # sub la_check

#----
# &la_check_old()
#
# Check load average
#----

sub la_check_old {
  # Glob the la logs
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu")) {
    (my $commahost = $log) =~ s/^.*\/|\.cpu$//g;
    (my $host = $commahost) =~ s/\,/\./g;

    my $ctime = (stat($log))[10];

    # Grab Load Avg line
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log file $log: $!");
      next;
    }

    my(@lines) = $infh->getlines();
    chomp(@lines);
    $infh->close;

    if ($lines[0] !~ /green|yellow|red/) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing red, yellow, or green. $log skipped.\n");
      next;
    }

    # Attempt to grab load average.
    # TODO: CPU usage and load average separately
    if (($lines[0] =~ /load\s*\=\s*([\d\.\%]+)/o) ||
        ($lines[0] =~ /CPU Usage\s*[\=\:]\s*([\d\.\%]+)/o)) {
      my $la = $1;

      if ($la =~ /^(\d+)\%$/) {
        # For bbnt client and routers
        $la = $1;
      } elsif ($la =~ /\./) {
        # If DISPREALLOADAVG="TRUE"
        $la = $la * 100;
      }

      print(localtime().": ".(caller(0))[3].": DEBUG la: $la\n") if $ENV{'DEBUG'};

      # RRD for la
  
      my $RRD="$ENV{'BBRRDS'}/$commahost.la.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.la.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }

      # if it doesn't exist, make it 
      if (-f $RRD ) {
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        } else {
          RRDs::update("$RRD",'-t','la',"$ctime:$la");
        }
      } else {
        # RRAS defined in larrd.pl
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"DS:la:GAUGE:600:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        RRDs::update("$RRD",'-t','la',"N:$la");
      }

      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }

    } elsif ($lines[2] =~ /% Processor Time:\s+([\.\d]+)/) {
      # Vladimir(coworker)'s custom BBSend VBScript
      # Does CPU usage as a percentage only
      # Use la as the service but use a different DS
      # His script runs only once every ten minutes because
      # it takes a full minute to run, so use a longer heartbeat.
      # That's the main reason for having a separate elsif
      # TODO: This is not the same as the load average one sees with the
      # uptime command on a UNIX system. This should be handled differently.

      my $cputime = $1;

      my $RRD="$ENV{'BBRRDS'}/$commahost.la.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.la.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }

      # if it doesn't exist, or if it has the wrong DS, make it 
      # If not, delete the bad RRD
      if ((-f $RRD ) &&
          ((exists(${&get_dsnames($RRD)}{'ProcessorTime'})) ||
           (! unlink($RRD)))) {
        if ($ctime == RRDs::last($RRD)) {
          print(localtime().": ".(caller(0))[3].": DEBUG $RRD has already been updated from $log\n") if $ENV{'DEBUG'};
          next;
        } else {
          RRDs::update("$RRD",'-t','ProcessorTime',"$ctime:$cputime");
        }
      } else {
        # RRAS defined in larrd.pl
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,"DS:ProcessorTime:GAUGE:800:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        RRDs::update("$RRD",'-t','ProcessorTime',"N:$cputime");
      }  # if (-f $RRD )
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }

    } elsif ($lines[0] =~ /CPU usage\:\s+(\d+)\%[^\,]+\,\s*(\d+)\%/i) { # elsif ($lines[2] =~ /% Processor Time:\s+([\.\d]+)/)
      # CPU usage as a pair of percentages for 1 minute and 5 minute averages
      my($cpu1min,$cpu5min) = ($1,$2);
      # TODO: Data storage and graphs for this format of CPU usage.
      # It would be a mistake to handle it the same as generic load average
      # that you see with the uptime command.

    } # elsif ($line =~ /CPU usage\:\s+(\d+)\%[^\,]+\,\s*(\d+)\%/i)
  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.cpu"))

  return 1;
} # sub la_check_old


#----
# &memory_check();
#
# Check memory usage, requires bb-memory.sh be installed as an extension.
#----

sub memory_check {
  # Get list of all memory log files
  # Loop through the list
  my %host2log = ();

  foreach my $logfile (&glob("$ENV{'BBLOGS'}/*.memory"),&glob("$ENV{'BBVAR'}/data/*.memory")) {
    (my $commahost = $logfile) =~ s/^.*\/|\.memory$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $logfile;
  }
  while (my($host,$logfile) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;

    # Get RRD file names
    my $realrrd = "$ENV{'BBRRDS'}/$commahost.memory.real.rrd";
    my $swaprrd = "$ENV{'BBRRDS'}/$commahost.memory.swap.rrd";

    # Get log inode change time
    my $ctime = (stat($logfile))[10];

    # Check if RRDs exists. If not, create them.
    # If they do, check that I am not duplicating timestamps
    if (-f $realrrd) {
      if ($ctime == RRDs::last($realrrd)) {
        next;
      }
    } else {
      RRDs::create($realrrd,'DS:realmempct:GAUGE:600:0:U',split(" ",$ENV{'RRAS'}));
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR Error creating RRD $realrrd: ".RRDs::error."\n");
        next;
      }
    } # if (-f $realrrd)
    if (-f $swaprrd) {
      if ($ctime == RRDs::last($swaprrd)) {
        next;
      }
      # Try to rename the DS if it has the wrong name
#      my $ds_ref = &get_dsnames($swaprrd);
#      if ((exists($ds_ref->{'realmempct'})) && (! exists($ds_ref->{'swapmempct'}))) {
#        RRDs::tune($RRD,'-r','realmempct:swapmempct'); # Data source rename
#        if (RRDs::error) {
#          print(localtime().": ".(caller(0))[3].": ERROR renaming data source 'realmempct' to 'swapmempct' in $RRD: ".RRDs::error."\n");
#          next;
#        }
#      }
    } else {
      # RRDs::create($swaprrd,'DS:swapmempct:GAUGE:600:0:U',split(" ",$ENV{'RRAS'}));
      RRDs::create($swaprrd,'DS:realmempct:GAUGE:600:0:U',split(" ",$ENV{'RRAS'}));
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR Error creating RRD $swaprrd: ".RRDs::error."\n");
        next;
      }
    } # if (-f $swaprrd)

    # Get Data:
    # Take the top three lines,
    # keep the lines that have the words Physical or Real,
    # take the fifth column,
    # remove the percent sign.
    my $infh = new FileHandle;
    unless ($infh->open("< $logfile")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open logfile $logfile: $!\n");
      next;
    }
    my($realpct,$swappct) = (0,0);
    while (my $line = $infh->getline()) {
      #              Name             Used    Total   Percentage
      if ($line =~ /(Physical|Real)\s+(\S+)\s+(\S+)\s+(\S+)/) {
        my($name,$used,$total,$percentage) = ($1,$2,$3,$4);
        $percentage =~ s/[^\d\.]//g;
        $realpct = $percentage;
      #                   Name         Used    Total   Percentage
      } elsif ($line =~ /(Swap|Page)\s+(\S+)\s+(\S+)\s+(\S+)/) {
        my($name,$used,$total,$percentage) = ($1,$2,$3,$4);
        $percentage =~ s/[^\d\.]//g;
        $swappct = $percentage;
      }
    }
    $infh->close;

    # Update RRD
    RRDs::update($realrrd,'-t','realmempct',"$ctime:$realpct");
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to update RRD file $realrrd with $realpct: ".RRDs::error."\n");
      next;
    }
    # RRDs::update($swaprrd,'-t','swapmempct',"$ctime:$swappct");
    RRDs::update($swaprrd,'-t','realmempct',"$ctime:$swappct");
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to update RRD file $swaprrd with $swappct: ".RRDs::error."\n");
      next;
    }
  } # foreach my $logfile (@memlogs)

  # Look for .cpu log files that contain NT memory data
  # Loop through the list
  %host2log = ();
  foreach my $logfile (&glob("$ENV{'BBLOGS'}/*.cpu"),&glob("$ENV{'BBVAR'}/data/*.cpu")) {
    (my $commahost = $logfile) =~ s/^.*\/|\.cpu$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $logfile;
  }
  while (my($host,$logfile) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;
    my $ctime = (stat($logfile))[10];

    # See if I have the PhysicalMem line that marks this as an NT CPU file
    my $infh = new FileHandle;
    unless ($infh->open("< $logfile")) {
      print(localtime().": ".(caller(0))[3].": ERROR Couldn't open logfile $logfile: $!\n");
      next;
    }
    my @cpulines = $infh->getlines();
    $infh->close();
    if ($cpulines[0] =~ /PhysicalMem: \d+\wB\((\d+)\%\)/) {
      my $realpct = $1;

      # Do the realpct first, since the necessary info for swap might
      # not be present if the client is too old.
      # Get RRD file name
      my $realrrd = "$ENV{'BBRRDS'}/$commahost.memory.real.rrd";
      # Check if RRD exists. If not, create it.
      if (-f $realrrd) {
        if ($ctime == RRDs::last($realrrd)) {
          next;
        }
      } else {
        RRDs::create($realrrd,'DS:realmempct:GAUGE:600:0:U',split(" ",$ENV{'RRAS'}));
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR Error creating RRD $realrrd: ".RRDs::error."\n");
          next;
        }
      } # if (-f $realrrd)

      # Update RRD
      RRDs::update($realrrd,'-t','realmempct',"$ctime:$realpct");
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR Unable to update RRD file $realrrd with $realpct: ".RRDs::error."\n");
        next;
      }

      # Get pagefile percentage if possible, use that as swap
      my($pagetotal) = (grep(/Total PageFile size/,@cpulines))[0] =~ /(\d+) bytes/;
      next unless $pagetotal;
      my($pageavail) = (grep(/Available PageFile size/,@cpulines))[0] =~ /(\d+) bytes/;
      my $swappct = int(100 * ($pagetotal - $pageavail) / $pagetotal);

      # Get RRD file name
      my $swaprrd = "$ENV{'BBRRDS'}/$commahost.memory.swap.rrd";
      # Check if RRD exists. If not, create it.
      if (-f $swaprrd) {
        if ($ctime == RRDs::last($swaprrd)) {
          next;
        }
        # Try to rename the DS if it has the wrong name
#        my $ds_ref = &get_dsnames($swaprrd);
#        if ((exists($ds_ref->{'realmempct'})) && (! exists($ds_ref->{'swapmempct'}))) {
#          RRDs::tune($RRD,'-r','realmempct:swapmempct'); # Data source rename
#          if (RRDs::error) {
#            print(localtime().": ".(caller(0))[3].": ERROR renaming data source 'realmempct' to 'swapmempct' in $RRD: ".RRDs::error."\n");
#            next;
#          }
#        }
      } else {
        # RRDs::create($swaprrd,'DS:swapmempct:GAUGE:600:0:U',split(" ",$ENV{'RRAS'}));
        RRDs::create($swaprrd,'DS:realmempct:GAUGE:600:0:U',split(" ",$ENV{'RRAS'}));
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR Error creating RRD $swaprrd: ".RRDs::error."\n");
          next;
        }
      } # unless (-f $swaprrd)

      # Update RRD
      # RRDs::update($swaprrd,'-t','swapmempct',"$ctime:$swappct");
      RRDs::update($swaprrd,'-t','realmempct',"$ctime:$swappct");
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR Unable to update RRD file $swaprrd with $swappct: ".RRDs::error."\n");
        next;
      }
    } else {
      next;
    }
  } # foreach my $logfile (glob("$ENV{'BBLOGS'}/*.cpu"))

  return 1;
} # sub memory_check


#----
# &netstat_check()
#
# Netstat check
#----

sub netstat_check {

  my %htonet = (
    solaris => {
      udpInDatagrams    => ['DERIVE','udpInDatagrams'],
      udpOutDatagrams   => ['DERIVE','udpOutDatagrams'],
      udpInErrors       => ['DERIVE','udpInErrors'],
      tcpActiveOpens    => ['DERIVE','tcpActiveOpens'],
      tcpPassiveOpens   => ['DERIVE','tcpPassiveOpens'],
      tcpAttemptFails   => ['DERIVE','tcpAttemptFails'],
      tcpEstabResets    => ['DERIVE','tcpEstabResets'],
      tcpCurrEstab      => ['GAUGE'  ,'tcpCurrEstab'],
      tcpOutDataBytes   => ['DERIVE','tcpOutDataBytes'],
      tcpInInorderBytes => ['DERIVE','tcpInInorderBytes'],
      tcpInUnorderBytes => ['DERIVE','tcpInUnorderBytes'],
      tcpRetransBytes   => ['DERIVE','tcpRetransBytes']
    },
    redhat => {
      "packets received"            => ['DERIVE','udpInDatagrams'],
      "packets sent"                => ['DERIVE','udpOutDatagrams'],
      "packet receive errors"       => ['DERIVE','udpInErrors'],
      "active connections openings" => ['DERIVE','tcpActiveOpens'],
      "passive connection openings" => ['DERIVE','tcpPassiveOpens'],
      "failed connection attempts"  => ['DERIVE','tcpAttemptFails'],
      "connection resets received"  => ['DERIVE','tcpEstabResets'],
      "connections established"     => ['GAUGE'  ,'tcpCurrEstab'], 
      "segments received"           => ['DERIVE','tcpInInorderPkts'],
      "segments send out"           => ['DERIVE','tcpOutDataPkts'],
      "segments retransmited"       => ['DERIVE','tcpRetransPkts']
    },
    hpux => {
      tcpPassiveOpens   => ['DERIVE','tcpPassiveOpens'],
      tcpCurrEstab      => ['GAUGE'  ,'tcpCurrEstab'],
      tcpOutDataBytes   => ['DERIVE','tcpOutDataBytes'],
      tcpInInorderBytes => ['DERIVE','tcpInInorderBytes'],
      tcpInUnorderBytes => ['DERIVE','tcpInUnorderBytes'],
      tcpRetransBytes   => ['DERIVE','tcpRetransBytes']
    },
  );
    
  $htonet{'linux'} = $htonet{'redhat'};
  $htonet{'suse'} = $htonet{'redhat'};

  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.netstat"),&glob("$ENV{'BBVAR'}/data/*.netstat")) {
    (my $commahost = $log) =~ s/^.*\/|\.netstat$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $log;
  }
  while (my($host,$log) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;

    print(localtime().": ".(caller(0))[3].": DEBUG log=$log\n") if ($ENV{'DEBUG'});

    # 2. grab the timpstamp for rrd. 

    my $ctime = (stat($log))[10];

    print(localtime().": ".(caller(0))[3].": DEBUG ctime = $ctime\n") if ($ENV{'DEBUG'});

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Couldn't open log '$log': $!");
      next;
    }
    my $temp = $infh->getline() if $log =~ /$ENV{'BBLOGS'}/io;
    my $bbhosttype = $infh->getline();
    chomp $bbhosttype;

    print(localtime().": ".(caller(0))[3].": DEBUG bbhosttype = $bbhosttype\n") if ($ENV{'DEBUG'});
    # 3. Grab the data from the log how ever you need

    # Checking for existence prevents perl from autovivifying the %htonet hash

    my(%hash)=();
    if($bbhosttype eq "solaris") {
      while (defined(my $netstat = $infh->getline()) ) {
        $netstat =~ /\w*\s+(\w+)\s*\=\s*(-)?(\d+)\s*((\w+)\s*\=\s*(-)?(\d+))?/o;
        if(defined $1) {
          print(localtime().": ".(caller(0))[3].": DEBUG 1=$1 2=$2 3=$3\n") if ($ENV{'DEBUG'});
          $hash{$1} = "$2$3" if exists $htonet{$bbhosttype}{$1};
        }
        if(defined $4) {
          print(localtime().": ".(caller(0))[3].": DEBUG 5=$5 6=$6 7=$7\n") if ($ENV{'DEBUG'});
          $hash{$5} = "$6$7" if exists $htonet{$bbhosttype}{$5};
        }
      }
    } elsif (($bbhosttype eq "redhat") ||
             ($bbhosttype eq "linux") ||
             ($bbhosttype eq "suse")) {
      my $section = '';
      while (defined(my $netstat = $infh->getline()) ) {
        if ($netstat =~ /^(\w):$/) {
          $section = $1;
          chomp($section);
          next;
        }
        $netstat =~ /\s*(-)?(\d+)\s*((\w+\s+)*\w+)/o;
        if (defined $2) {
          print(localtime().": ".(caller(0))[3].": DEBUG Section=$section 1=$1 2=$2 3=$3\n") if ($ENV{'DEBUG'});
          $hash{$htonet{$bbhosttype}{$3}[1]} = "$1$2" if exists $htonet{$bbhosttype}{$3};
        }
      }
    } elsif ($bbhosttype eq "hpux") {
      while (defined(my $netstat = $infh->getline()) ) {
        $netstat =~ /(\w+)\s+\=\s+(-)?(\d+)/o;
        if (defined($1)) {
          print(localtime().": ".(caller(0))[3].": DEBUG 1=$1 2=$2 3=$3\n") if ($ENV{'DEBUG'});
          $hash{$1} = "$2$3" if exists $htonet{$bbhosttype}{$1};
        }
      } # while (defined(my $netstat = $infh->getline()) ) {
    } else {
      print(localtime().": ".(caller(0))[3].": WARN bbhostype undetermined : $bbhosttype. Ignoring host $host.\n");
      next;
    }
      
    # 4. create the rrd filename

    my $RRD="$ENV{'BBRRDS'}/$commahost.netstat.rrd";
    my $oldRRD="$ENV{'BBRRDS'}/$host.netstat.rrd";
    if (($RRD ne $oldRRD) &&
        (-f $oldRRD)) {
      unless (rename($oldRRD,$RRD)) {
        print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
      }
    }

    print(localtime().": ".(caller(0))[3].": RRD=$RRD\n") if ($ENV{'DEBUG'});

    # see if the RRD is around, if not make it  
    if ( not -f $RRD ) {
      my $start = '';
      # build the @ds
      my @ds =();
      foreach (values %{$htonet{$bbhosttype}}) {
        $start = ($_->[0] eq "GAUGE") ? "1" : "0";
        push @ds,"DS:$_->[1]:$_->[0]:600:$start:U";
      }  

      # keeping this much disk data seems overkill
      # but thought 'standardizing' RRD's would be worth it
      # 30 seconds....
      $ctime--;
      my @rras = split " ",$ENV{'RRAS'};
      RRDs::create($RRD,"-b $ctime",@ds,@rras);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
        next;
      }
      $ctime++;
      print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
    }

    # 5. Put your data from step 2 to the rrd

    # now update the RRD
  
    my($rrdupdate)="$ctime";
    my $rrdupdate_t = '';

    # The original assumes that 'values' will return elements in the same order
    # both times it is called. This is probably true but not
    # necessarily true.
    foreach my $dsa_ref (values %{$htonet{$bbhosttype}}) {
      print(localtime().": ".(caller(0))[3].": ".$dsa_ref->[1]." = ".$hash{$dsa_ref->[1]}."\n") if ($ENV{'DEBUG'});
      $rrdupdate .= ':'.$hash{$dsa_ref->[1]};
      $rrdupdate_t .= $dsa_ref->[1].':';
    }
    chop($rrdupdate_t);

    print(localtime().": ".(caller(0))[3].": DEBUG rrdupate=$rrdupdate\n") if ($ENV{'DEBUG'});

    RRDs::update($RRD,'-t',$rrdupdate_t,$rrdupdate);
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": WARN updating $RRD: '".RRDs::error."', order '$rrdupdate_t', data '$rrdupdate'\n");
    }
    &unlink_data($log);
  } # while (my($host,$log) = each(%host2log))
  return 1;
} # sub netstat_check


#----
# &sendmail_check();
#
# Sendmail check
#----

sub sendmail_check {

  my @ds = ("DS:msgsfr:DERIVE:600:0:U",
       "DS:bytes_from:DERIVE:600:0:U",
       "DS:msgsto:DERIVE:600:0:U",
       "DS:bytes_to:DERIVE:600:0:U",
       "DS:msgsrej:DERIVE:600:0:U",
       "DS:msgsdis:DERIVE:600:0:U");
  my $dsorder = 'msgsfr:bytes_from:msgsto:bytes_to:msgsrej:msgsdis';

  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.sendmail"),&glob("$ENV{'BBVAR'}/data/*.sendmail")) {
    (my $commahost = $log) =~ s/^.*\/|\.sendmail$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $log;
  }
  while (my($host,$log) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;

    print(localtime().": ".(caller(0))[3].": DEBUG log=$log\n") if ($ENV{'DEBUG'});
    my $ctime = (stat($log))[10];
    print(localtime().": ".(caller(0))[3].": DEBUG ctime = $ctime\n") if ($ENV{'DEBUG'});

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Couldn't open logfile '$log': $!");
      next;
    }
  
    my $temp = $infh->getline() if $log =~ /$ENV{'BBLOGS'}/io;

    while(defined(my $sendmail = $infh->getline())) {
      next if $sendmail =~ /^Statistics/;
      my $new_sm = 1 if ($sendmail =~ /msgsrej msgsdis/);
      next if $sendmail =~ /Mailer$/;
      last if $sendmail =~ /^\=/;
      chomp $sendmail;
      my($M,$mfr,$bfr,$mt,$bt,$mr,$md,$mailer);
      if ($new_sm) {
        ($M,$mfr,$bfr,$mt,$bt,$mr,$md,$mailer) = split " ", $sendmail;
      } else {
        ($M,$mfr,$bfr,$mt,$bt,$mailer) = split " ", $sendmail;
      }

      my $commahost = $host;
         $commahost =~ s/\./\,/g;
      my $RRD="$ENV{'BBRRDS'}/$commahost.sendmail.$mailer.rrd";
      my $oldRRD="$ENV{'BBRRDS'}/$host.sendmail.$mailer.rrd";
      if (($RRD ne $oldRRD) &&
          (-f $oldRRD)) {
        unless (rename($oldRRD,$RRD)) {
          print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
        }
      }

      print(localtime().": ".(caller(0))[3].": DEBUG RRD = $RRD\n") if ($ENV{'DEBUG'});

      # see if the RRD is around, if not make it  
      if ( not -f $RRD ) {
      
        # keeping this much disk data seems overkill
        # but thought 'standardizing' RRD's would be worth it
        # 30 seconds....
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($RRD,@ds,@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
          next;
        }
        $ctime="N";
        print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
      }

      my($rrdupdate)="$ctime:$mfr:$bfr:$mt:$bt:$mr:$md";

      print(localtime().": ".(caller(0))[3].": DEBUG rrdupdate = $rrdupdate\n") if ($ENV{'DEBUG'});

      # now update the RRD
      RRDs::update($RRD,'-t',$dsorder,$rrdupdate);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
      }
    }
    $infh->close;
    &unlink_data($log);
  } # while (my($host,$log) = each(%host2log))
  return 1;
} # sub sendmail_check


#----
# &vmstat_check()
#
# Check vmstat
#----

sub vmstat_check {
# Define some host vmstat hash column translation

  my %htovm = (
    solaris => {
      cpu_r     => 0,
      cpu_b     => 1,
      cpu_w     => 2,
      mem_swap  => 3,
      mem_free  => 4,
      mem_re    => 5,
      mem_mf    => 6,
      mem_pi    => 7,
      mem_po    => 8,
      sr        => 11,
      cpu_int   => 16,
      cpu_syc   => 17,
      cpu_csw   => 18,
      cpu_usr   => 19,
      cpu_sys   => 20,
      cpu_idl   => 21,
    },
    redhat => {
      cpu_r    => 0,
      cpu_b    => 1,
      cpu_w    => 2,
      mem_swpd => 3,
      mem_free => 4,
      mem_buff => 5,
      mem_cach => 6,
      mem_si   => 7,
      mem_so   => 8,
      dsk_bi   => 9,
      dsk_bo   => 10,
      cpu_int  => 11,
      cpu_csw  => 12,
      cpu_usr  => 13,
      cpu_sys  => 14,
      cpu_idl  => 15,
    },
    osf => {
      cpu_r    => 0, 
      cpu_b    => 1, 
      cpu_w    => 2, 
      mem_free => 4, 
      mem_mf   => 6, 
      mem_pi   => 10, 
      mem_po   => 11, 
      cpu_int  => 12, 
      cpu_syc  => 13, 
      cpu_csw  => 14, 
      cpu_usr  => 15, 
      cpu_sys  => 16, 
      cpu_idl  => 17, 
    }, 
    aix     => {
      cpu_r    => 0,
      cpu_b    => 1,
      mem_avm  => 2,
      mem_free => 3,
      mem_re   => 4,
      mem_pi   => 5,
      mem_po   => 6,
      mem_fr   => 7,
      sr       => 8,
      mem_cy   => 9,
      cpu_int  => 10,
      cpu_syc  => 11,
      cpu_csw  => 12,
      cpu_usr  => 13,
      cpu_sys  => 14,
      cpu_idl  => 15,
      cpu_wait => 16,
    },
    freebsd => {
      cpu_r    => 0,
      cpu_b    => 1,
      cpu_w    => 2,
      mem_avm  => 3,
      mem_free => 4,
      mem_flt  => 5,
      mem_re   => 6,
      mem_pi   => 7,
      mem_po   => 8,
      mem_fr   => 9,
      sr       => 10,
      dsk_da0  => 11,
      dsk_fd0  => 12,
      cpu_int  => 13,
      cpu_csw  => 15,
      cpu_sys  => 16,
      cpu_usr  => 17,
      cpu_idl  => 18,
    },
  );

  # Aliases
  $htovm{'debian'}=$htovm{'redhat'};
  $htovm{'linux'}=$htovm{'redhat'};
  $htovm{'suse'}=$htovm{'redhat'};

  my %host2log = ();
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.vmstat"),&glob("$ENV{'BBVAR'}/data/*.vmstat")) {
    (my $commahost = $log) =~ s/^.*\/|\.vmstat$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $host2log{$host} = $log;
  }
  while (my($host,$log) = each(%host2log)) {
    (my $commahost = $host) =~ s/\./\,/g;

    print(localtime().": ".(caller(0))[3].": DEBUG log=$log\n") if ($ENV{'DEBUG'});
    my $ctime = (stat($log))[10];
    print(localtime().": ".(caller(0))[3].": DEBUG ctime = $ctime\n") if ($ENV{'DEBUG'});
    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Couldn't open logfile '$log': $!");
      next;
    }
    # Siphon off the false status line if I'm not allowing data messages
    my $temp = $infh->getline() if $log =~ /$ENV{'BBLOGS'}/io;
    my $bbhosttype = $infh->getline();
    chomp $bbhosttype;
    unless (exists($htovm{$bbhosttype})) {
      print(localtime().": ".(caller(0))[3].": WARN Unknown host type '$bbhosttype' for host '$host' in logfile '$log'\n");
      next;
    }

    print(localtime().": ".(caller(0))[3].": DEBUG bbhosttype = $bbhosttype\n") if ($ENV{'DEBUG'});
    my $vmstat = $infh->getline();
    print(localtime().": ".(caller(0))[3].": DEBUG vmstat = $vmstat\n") if ($ENV{'DEBUG'});
    $infh->close;

    chomp $vmstat;
    my @cols = split " ",$vmstat;

    my $RRD="$ENV{'BBRRDS'}/$commahost.vmstat.rrd";
    my $oldRRD="$ENV{'BBRRDS'}/$host.vmstat.rrd";
    if (($RRD ne $oldRRD) &&
        (-f $oldRRD)) {
      unless (rename($oldRRD,$RRD)) {
        print(localtime().": ".(caller(0))[3].": ERROR Failed to rename old style RRD filename '$oldRRD' to newer style filename '$RRD': $!\n");
      }
    }
    print(localtime().": ".(caller(0))[3].": DEBUG RRD = $RRD\n") if ($ENV{'DEBUG'});

    # See if the RRD is around. If so, make sure it has the right
    # datasources. If it doesn't, delete it.
    # Datasources stored in the RRD in alphabetical order, so no point in
    # checking the order.
    if (-f $RRD) {
      if (my $rrdds_ref = &get_dsnames($RRD)) {
        # Create modified version of %htovm without the missing column numbers
        my @dsnames = sort { $htovm{$bbhosttype}{$a} <=> $htovm{$bbhosttype}{$b} } keys(%{$htovm{$bbhosttype}});
        my $n = 0;
        my %newhtovm = map { $_ => $n++ } @dsnames;
        # %newhtovm and %$rrdds_ref should be equivalent
        while (my $dsname = each(%newhtovm)) {
          unless (exists($rrdds_ref->{$dsname})) {
            print(localtime().": ".(caller(0))[3].": WARN data source '$dsname' found in htovm but not in RRD '$RRD'\n");
            unless (unlink($RRD)) {
              print(localtime().": ".(caller(0))[3].": ERROR Unable to remove $RRD with wrong DS names\n");
            }
            last;
          }
        } # while (my $dsname = each(%newhtovm))

        while (my $dsname = each(%$rrdds_ref)) {
          last unless (-f $RRD);
          unless (exists($newhtovm{$dsname})) {
            print(localtime().": ".(caller(0))[3].": WARN data source '$dsname' found in RRD '$RRD' but not in htovm\n");
            unless (unlink($RRD)) {
              print(localtime().": ".(caller(0))[3].": ERROR Unable to remove $RRD with wrong DS names\n");
            }
            last;
          }
        } # while (my $dsname = each(%$rrdds_ref))
      } else {
        print(localtime().": ".(caller(0))[3].": ERROR Failed to get DS names from RRD file $RRD for host $host\n");
      }
    } # if (-f $RRD)

    my(@ds);
    my $dsorder = '';
    my @dsnames = ();
    foreach my $col (sort(keys(%{$htovm{$bbhosttype}}))) {
      push @ds,"DS:$col:GAUGE:600:0:U";
      push(@dsnames,$col);
      $dsorder .= "$col:";
    }
    chop($dsorder);

    # see if the RRD is not around, if not make it  
    if (! -f $RRD) {
      print(localtime().": ".(caller(0))[3].": DEBUG ds = @ds\n") if ($ENV{'DEBUG'});
      my @rras = split " ",$ENV{'RRAS'};
      RRDs::create($RRD,@ds,@rras);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error.": ".join(' ',$RRD,@ds,@rras)."\n");
        next;
      }
      $ctime="N";
      print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
    }

    my($rrdupdate)="$ctime";
    foreach my $col (@dsnames) { 
      $rrdupdate .= ":$cols[$htovm{$bbhosttype}{$col}]";
    }
    print(localtime().": ".(caller(0))[3].": DEBUG rrdupdate = $rrdupdate\n") if ($ENV{'DEBUG'});

    # now update the RRD
    RRDs::update($RRD,'-t',$dsorder,$rrdupdate);
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": WARN updating $RRD: ".RRDs::error."\n");
    }
    &unlink_data($log);
  } # while (($host,$log) = each(%h_fn))

  return 1;
} # sub vmstat_check


#----
# &mailq_check()
#
# Mail queue checks.
# Compatible with Pascal Peloni's bb-mailq.sh v1.1 that can handle sendmail and
# James Overbeck's bb-mailqueue.sh v1.3 that can handle either sendmail or qmail.
#----

sub mailq_check {

  # Glob the mailq logs
  my @rras = split " ",$ENV{'RRAS'};
  foreach my $log (&glob("$ENV{'BBLOGS'}/*.mailq")) {
    (my $commahost = $log) =~ s/^.*\/|\.mailq$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    my $ctime = (stat($log))[10];

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log '$log': $!");
      next;
    }
    my @loglines = $infh->getlines();
    $infh->close;
    chomp(@loglines);

    if ($loglines[0] !~ /^green |^yellow |^red /) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing red, yellow, or green. $log skipped.\n");
      next;
    }

    my $mailcount = undef;
    my $mailspace = undef; # In bytes

    # Try to extract data from log contents

    if ((scalar(grep(/Mail queue contains\s+\d+\s+item/,@loglines))) &&
             (scalar(grep(/Mail queue size/,@loglines)))) {
      # Do I have a count and total size in the body
      # like bb-mailq.sh v1.1 and bb-mailqueue.sh v1.3?
      ($mailcount) = (grep(/Mail queue contains\s+\d+\s+item/,@loglines))[0] =~ /Mail queue contains\s+(\d+)\s+item/;
      my $spacelabel = '';
      ($mailspace,$spacelabel) = (grep(/Mail queue size/,@loglines))[0] =~ /Mail queue size\s*\:\s*(\d+)\s*([A-Za-z]+)/;
      if ($spacelabel =~ /^kb$/i) { # kilo
        $mailspace = 1024 * $mailspace;
      } elsif ($spacelabel =~ /^mb$/i) { # mega
        # Lots of mail
        $mailspace = 1024 * 1024 * $mailspace;
      } elsif ($spacelabel =~ /^gb$/i) { # giga
        # A ton of mail
        $mailspace = 1024 * 1024 * 1024 * $mailspace;
      } elsif ($spacelabel =~ /^tb$/i) { # tera
        # Spam attack
        $mailspace = 1024 * 1024 * 1024 * 1024 * $mailspace;
      }
    } elsif ($loglines[0] =~ /(\d+)\s+requests/) {
      # Do I have number of messages in the first line like
      # larrd-0.43a expects?
      $mailcount = $1;
    } # elsif ($loglines[0] =~ /(\d+)\s+requests/)

    # Create/update RRD files based on data found.

    if (defined($mailcount)) {
      my $RRD = "$ENV{'BBRRDS'}/$commahost.mailq.rrd";
      if (defined($mailspace)) {
        # bb-mailq.sh style: Count and size
        if (-f $RRD) {
          # TODO:
          # See if the RRD is a count-only style.
          # If so, use RRDs::tune to change it.
        } else {
          RRDs::create($RRD,"DS:mailq:GAUGE:600:0:U","DS:mailspace:GAUGE:600:0:U",@rras);
          if (RRDs::error) {
            print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
            next;
          }
          $ctime = 'N';
        } # if (-f $RRD)
        RRDs::update($RRD,'-t','mailq:mailspace',"$ctime:$mailcount:$mailspace");
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR updating $RRD: ".RRDs::error."\n");
        }
      } else { # if (defined($mailspace))
        # Old style: Count only
        if (! -f $RRD) {
          RRDs::create($RRD,"DS:mailq:GAUGE:600:0:U",@rras);
          if (RRDs::error) {
            print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
            next;
          }
          $ctime = 'N';
          print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
        } # if (! -f $RRD)
        RRDs::update($RRD,'-t','mailq',"$ctime:$mailcount");
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR updating $RRD: ".RRDs::error."\n");
        }
      } # if (defined($mailspace))
    }  else { # if (defined($mailcount))
      # We have a problem: not even a count of the number of messages was found
      print(localtime().": ".(caller(0))[3].": ERROR no graphable data found in $log\n");
    } # if (defined($mailcount))

  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.mailq"))
  return 1;
} # sub mailq_check


#----
# &temperature_check();
#
# Temperature checks.
# Immitates larrd-0.43c.
# Handles SunEnterprise/SunFire/etc from temperature.sh
# Handles routers temperatures from bb-xsnmptemperature.pl
#----

sub temperature_check {
  my $ds = 'temperature';

  foreach my $log (&glob("$ENV{'BBLOGS'}/*.$ds")) {
    (my $commahost = $log) =~ s/^.*\/|\.$ds$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    my $ctime = (stat($log))[10];

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log '$log': $!");
      next;
    }
    my @loglines = $infh->getlines();
    $infh->close;
    chomp(@loglines);

    if ($loglines[0] !~ /^green |^yellow |^red /) {
      print(localtime().": ".(caller(0))[3].": WARN $host is not showing red, yellow, or green. $log skipped.\n");
      next;
    }

    if (scalar(grep(/^Device\s+Temp\(C\)\s+Temp\(F\)/,@loglines))) { # look for the table header
      # temperature.sh
      # based on larrd-0.43c. Don't understand everything done there so
      # mostly copied. Some regexps changed.
      foreach my $line (@loglines) {
        # Change / to , and single spaces to _
        $line =~ s/\//,/g;
        $line =~ s/(\S) (\S)/$1_$2/g;
        # Do space check again in case of "X Y Z" matches
        $line =~ s/(\S) (\S)/$1_$2/g;
        # Removed any HTML tags. Forget multiline tags.
        $line =~ s/\<[^\>]+\>//g;
        $line =~ s/\&red_*//g;
        $line =~ s/\&yellow_*//g;
        $line =~ s/\&green_*//g;
        print(localtime().": ".(caller(0))[3].": DEBUG $line\n") if $ENV{'DEBUG'};
        my @col = split(" ",$line);
        # Throw away lines that are not data
        next if ($col[1] !~ /\d+/); # Skip if column 2 has no numbers
        next if ($col[2] !~ /\d+/); # Skip if column 3 has no numbers
        next if ($line =~ /Temp\(F\)/); # Skip the header line, and this is where larrd 0.43c does the regexp wrong
        next if ($col[0] =~ /^green|^yellow|^red|^purple|^clear|^blue/); # skip the status line
        next if ($col[0] =~ /\d+.\d+_C/); # Skip if the first column has a temperature. It should have a device name.

        # create the rrd filename
        my $RRD = "$ENV{'BBRRDS'}/$host.$ds,$col[0].rrd";
        my $temperature = $col[1];
        print(localtime().": ".(caller(0))[3].": RRD=$RRD\n") if $ENV{'DEBUG'};
        unless (&temperature_check_update($RRD,$temperature,$ctime)) {
          print(localtime().": ".(caller(0))[3].": ERROR couldn't add temperature $temperature ctime $ctime to RRD file $RRD\n");
          next;
        }
      } # foreach my $line (@loglines)
    } elsif (scalar(grep(/bb-xsnmp/,@loglines))) { # look for the credits
      # bb-xsnmptemperature.pl
      next if scalar(grep(/Network Appliance/,@loglines)); # NetApps just say if they're hotornot
      if (my($line) = grep(/Actual Temperature\s*\:\s*\d+/,@loglines)) {
        # Foundry
        if ($line =~ /Actual Temperature\s*\:\s*(\d+)/) {
          my $temperature = $1;
          my $RRD = "$ENV{'BBRRDS'}/$commahost.$ds,Actual.rrd";
          print(localtime().": ".(caller(0))[3].": RRD=$RRD\n") if $ENV{'DEBUG'};
          unless (&temperature_check_update($RRD,$temperature,$ctime)) {
            print(localtime().": ".(caller(0))[3].": ERROR couldn't add temperature $temperature ctime $ctime to RRD file $RRD\n");
            next;
          } # unless (&temperature_check_update($RRD,$temperature,$ctime))
        } # if ($line =~ /Actual Temperature\s*\:\s*(\d+)/)
      } elsif (my(@templines) = grep(/^\&?(green|yellow|red)\s+.+C\, Threshold\: \d+ .*\, State\: /,@loglines)) {
        # Cisco
        foreach my $line (@templines) {
          # Similar to larrd-0.43c
          if ($line =~ /^\&?(?:green|yellow|red)\s+([^\:]+)\:\s*(\d+)\D/) {
            my($name,$temperature) = ($1,$2);
            $name =~ s/^\s//g;
            $name =~ s/\s$//g;
            $name =~ s/\W/_/g;
            $name =~ s/__/_/g;

            # create the rrd filename
            my $RRD = "$ENV{'BBRRDS'}/$commahost.$ds,$name.rrd";
            print(localtime().": ".(caller(0))[3].": RRD=$RRD\n") if $ENV{'DEBUG'};
            unless (&temperature_check_update($RRD,$temperature,$ctime)) {
              print(localtime().": ".(caller(0))[3].": ERROR couldn't add temperature $temperature ctime $ctime to RRD file $RRD\n");
              next;
            }
          } else {
            print(localtime().": ".(caller(0))[3].": ERROR unknown format for cisco $host in logfile $log\n");
          }
        } # foreach my $line (@templines)
      } else {
        print(localtime().": ".(caller(0))[3].": ERROR unknown format for host $host in logfile $log\n");
        next;
      }
    } else {
      # Unknown format
      print(localtime().": ".(caller(0))[3].": WARN $host has unknown format. $log skipped.\n");
      next;
    }
  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.temperature"))

  return 1;
} # sub temperature_check


#----
# &temperature_check_update($RRD,$temperature,$ctime);
#
# Update temperature for a single DS rrd file
#----

sub temperature_check_update {
  my($RRD,$temperature,$ctime) = @_;
  my @dsources = ("DS:temperature:GAUGE:600:1:U");
  my @rras = split " ",$ENV{'RRAS'};

  if (-f $RRD) {
    if ($ctime == RRDs::last($RRD)) {
      # Already been updated. Return quietly.
      return 1;
    }
  } else {
    RRDs::create($RRD,@dsources,@rras);
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": ERROR creating $RRD: ".RRDs::error."\n");
      return;
    }
    $ctime = "N";
    print(localtime().": ".(caller(0))[3].": STATUS did not find $RRD, created.\n") if ($ENV{'STATUS'});
  } # if (! -f $RRD)
  # Now update
  RRDs::update("$RRD",'-t','temperature',"$ctime:$temperature");
  if (RRDs::error) {
    print(localtime().": ".(caller(0))[3].": ERROR updating $RRD: ".RRDs::error."\n");
    return;
  }

  return 1;
}


#----
# &files_check
#
# For use with bb-files from deadcat.net.
# Does a very simplistic check of files sizes for files whose sizes are checked.
# Creates graphs somewhat like disk_check graphs.
#----

sub files_check {
  my $ds = 'files';

  foreach my $log (&glob("$ENV{'BBLOGS'}/*.$ds")) {
    (my $commahost = $log) =~ s/^.*\/|\.$ds$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    my $ctime = (stat($log))[10];

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log '$log': $!\n");
      next;
    }
    my @loglines = $infh->getlines();
    $infh->close;
    chomp(@loglines);

    # First, cleanup: Make sure <TABLE> and </TABLE> lines are on their own
    for (my $n = 0; $n <= $#loglines; $n++) {
      if ($loglines[$n] =~ /^(.+)(<TABLE.+)$/i) {
        my($first,$second) = ($1,$2);
        splice(@loglines,$n,1,$first,$second);
      } elsif ($loglines[$n] =~ /^(.+)(<\/TABLE>.*)$/i) {
        my($first,$second) = ($1,$2);
        splice(@loglines,$n,1,$first,$second);
      } elsif ($loglines[$n] =~ /^(<TABLE [^>]+>)(.+)$/i) {
        my($first,$second) = ($1,$2);
        splice(@loglines,$n,1,$first,$second);
      } elsif ($loglines[$n] =~ /^(<\/TABLE>)(.+)$/i) {
        my($first,$second) = ($1,$2);
        splice(@loglines,$n,1,$first,$second);
      }
      # Prevent loops
      # If I get here I screwed up log cleanup
      if ($n > 100) {
        print(localtime().": ".(caller(0))[3].": ERROR more than 100 lines in '$log'\n");
        last;
      }
    }

    # Eliminate "should not exist" or "too many"
    # @loglines = grep(!/should not exist/,@loglines);
    # @loglines = grep(!/too many/,@loglines);
    # @loglines = grep(!/not found/,@loglines);
  
    my %filesizedb = ();
    my $is_intable = 0;
    my $is_insizetable = 0;
    # Find the table that has Status Filename Size Age Date as the header.
    # They need to be in that order or else it won't work.
    foreach my $line (@loglines) {
      if ($is_intable) {
        if ($is_insizetable) {
          if ($line =~ /<\/TABLE>/i) {
            $is_insizetable = 0;
            $is_intable = 0;
          } elsif ($line =~ /^<TR><TD>[^<]+<\/TD><TD>([^<]+)<\/TD><TD>([^<]+)<\/TD>/i) {
            my($filename,$filesize) = ($1,$2);
            # Skip the file if it has anything that looks like a wildcard
            next if $filename =~ /\*/;
            next if $filename =~ /\?/;
            # Assume file sizes in bytes
            $filesizedb{$filename} = $filesize;
          } else {
            print(localtime().": ".(caller(0))[3].": ERROR line '$line' for $host\n");
          }
        } elsif ($line =~ /Status.*Filename.*Size.*Age.*Date/) {
          $is_insizetable = 1;
        } else {
          # Oops wrong table
          $is_intable = 0;
        }
      } elsif ($line =~ /^\<TABLE/i) {
        $is_intable = 1;
      }
    } # foreach my $line (@loglines)

    if (scalar(keys(%filesizedb)) <= 0) {
      print(localtime().": ".(caller(0))[3].": WARN no file sizes found for $host\n");
      next;
    }

    # Create separate rrd file for each file.
    # Immitate disk_check
    while (my($filename,$filesize) = each(%filesizedb)) {
      (my $tempname = $filename) =~ s/\//\,/g;
      my $rrdfile = "$ENV{'BBRRDS'}/$commahost.$ds$tempname.rrd";

      # If it doesn't exist, make it.
      # If it does, don't update it twice for the same point in time.
      if (-f $rrdfile) {
        if ($ctime == RRDs::last($rrdfile)) {
            print(localtime().": ".(caller(0))[3].": DEBUG $rrdfile has already been updated from $log\n") if $ENV{'DEBUG'};
            next;
        }
      } else {
        my @rras = split " ",$ENV{'RRAS'};
        RRDs::create($rrdfile,"DS:size:GAUGE:600:0:U",@rras);
        if (RRDs::error) {
          print(localtime().": ".(caller(0))[3].": ERROR creating $rrdfile: ".RRDs::error."\n");
          next;
        }
        # New RRDs don't like old dates
        $ctime = "N";
        print(localtime().": ".(caller(0))[3].": STATUS did not find $rrdfile, created.\n") if ($ENV{'STATUS'});
      } # if (-f $rrdfile) 
      print("$rrdfile $ctime:$filesize\n") if ($ENV{'DEBUG'});
      RRDs::update("$rrdfile",'--template','size',"$ctime:$filesize");
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": WARN updating $rrdfile: ".RRDs::error."\n");
      }
    } # while (my($filename,$filesize) = each(%filesizedb))

  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.$ds"))
  return 1;
} # sub files_check


#----
# &ftpusers_check
#
# Counts the number of FTP users connected.
# Only works when the server gives that information during anonymous logins.
# Otherwise, silently skip.
#----

sub ftpusers_check {

  my %logdb = ();

  foreach my $log (&glob("$ENV{'BBLOGS'}/*.ftp"),&glob("$ENV{'BBVAR'}/data/*.ftpusers")) {
    (my $commahost = $log) =~ s/^.*\/|\.ftp(users)?$//g;
    (my $host = $commahost) =~ s/\,/\./g;
    $logdb{$host} = $log;
  }

  while (my($host,$log) = each(%logdb)) {
    (my $commahost = $log) =~ s/^.*\/|\.ftp(users)?$//g;

    my $ctime;
    if ($log eq "$ENV{'BBLOGS'}/$commahost.ftp") {
      $ctime = (stat($log))[10];
    }

    my $infh = new FileHandle;
    unless ($infh->open("< $log")) {
      print(localtime().": ".(caller(0))[3].": ERROR Unable to open log '$log': $!\n");
      next;
    }
    my @loglines = $infh->getlines();
    $infh->close;
    chomp(@loglines);

    # Proceed only if the number of users is given
    my $users = 'U';
    my $maxusers = 'U';

    if ($log eq "$ENV{'BBVAR'}/data/$commahost.ftpusers") {
      unlink($log);
      if ($loglines[-1] =~ /^(\d+)\s*\:\s*(\d+)/) { # last line
        $ctime = $1;
        $users = $2;
      } else {
        # Huh, something's wrong.
        print(localtime().": ".(caller(0))[3].": ERROR number of FTP users not found in $log\n");
        next;
      }
    } elsif (my($userline) = grep(/You are user number \d+ of \d+ allowed/,@loglines)) {
      if ($userline =~ /You are user number (\d+) of (\d+) allowed/) {
        $users = $1;
        $maxusers = $2;
        $users = $users - 1; # Minus my own connection
      }
    } else {
      # Number of connected users not given. Silently skip
      next;
    }

    my $rrdfile = "$ENV{'BBRRDS'}/$commahost.ftpusers.rrd";

    if (-f $rrdfile) {
      # TODO: Check maxusers

      if ($ctime == RRDs::last($rrdfile)) {
        print(localtime().": ".(caller(0))[3].": DEBUG $rrdfile has already been updated from $log\n") if $ENV{'DEBUG'};
        next;
      }
    } else {
      my @rras = split " ",$ENV{'RRAS'};
      RRDs::create($rrdfile,"DS:users:GAUGE:600:0:$maxusers",@rras);
      if (RRDs::error) {
        print(localtime().": ".(caller(0))[3].": ERROR creating $rrdfile: ".RRDs::error."\n");
        next;
      }
      # New RRDs don't like old dates
      $ctime = "N";
      print(localtime().": ".(caller(0))[3].": STATUS did not find $rrdfile, created.\n") if $ENV{'STATUS'};
    } # if (-f $rrdfile)

    print("$rrdfile $ctime:$users\n") if ($ENV{'DEBUG'});
    RRDs::update($rrdfile,'--template','users',"$ctime:$users");
    if (RRDs::error) {
      print(localtime().": ".(caller(0))[3].": WARN updating $rrdfile: ".RRDs::error."\n");
    }
  } # foreach my $log (&glob("$ENV{'BBLOGS'}/*.ftp"))

  return 1;
} # sub ftpusers_check


1;
